
#if SUPD==605

#ifndef _DAEMONS_PORTALCONNECT_HXX_
#define _DAEMONS_PORTALCONNECT_HXX_

#ifdef __cplusplus

#include <rtl/ustring>
#include <rtl/byteseq.hxx>
#include <vos/ref.hxx>


namespace daemons
{

#define COMPRESSED          "compressed"
#define SECURE              "secure"
#define COMPRESSED_SECURE   "compressed_secure"

//---------------------------------------------------------------------------
/** enumeration of error values
 */

enum RC
{
    E_None                      = 0,
    E_INVALID_TICKET            = 1,
    E_WRONG_PASSWORD            = 2,
    E_UNKNOWN_USER              = 3,
    E_INVALID_USER_OR_PASSWORD  = 4, 
    E_USER_ALREADY_LOGGED_IN    = 5,
    E_NO_SUCH_SERVICE           = 6,
    E_CONNECTION_FAILED         = 7,
    E_UNKNOWN                   = 8,
    E_ILLEGAL_ARGUMENT          = 9,
    E_UNKNOWN_HOST              = 10,
    E_TERMINATE_REQUESTED       = 11
};


//---------------------------------------------------------------------------
/** abstract connection class
 */
  
class IConnection: public ::vos::IReference
{
public:

  //---------------------------------------------------------------------------
  /** reads the specified amount of bytes from the stream. The call will block until
      nBytesToRead bytes are available.

	  @param rDataArra    [out]: the buffer to fill.
	  @param nBytesToRead [in] : the number of bytes to read.

	  @returns the number of bytes read;
  */

  virtual sal_Int32 SAL_CALL read( sal_Int8* pData, sal_Int32 nBytesToRead ) = 0;

 //---------------------------------------------------------------------------
  /** tries to read the specified amount of bytes from the stream.

	  @param rDataArra    [out]: the buffer to fill.
	  @param nBytesToRead [in] : the number of bytes to read.

	  @returns the number of bytes read;
  */

  virtual sal_Int32 SAL_CALL readSomeBytes( sal_Int8* pData, sal_Int32 nBytesToRead ) = 0;


  //---------------------------------------------------------------------------
  /** writes the specified number of bytes to the stream.

	  @param rDataArra    [in]: the buffer to fill.
   */

 virtual sal_Int32 SAL_CALL write( const sal_Int8* pData , sal_Int32 nBytesToWrite ) = 0;

  //---------------------------------------------------------------------------
  /** flushes all output data to the stream.
   */

  virtual sal_Int32 SAL_CALL flush() = 0;

  //---------------------------------------------------------------------------
  /** closes the stream of the connection.

	  NOTE: implementation must be able to handle more than one close calls.
   */
  virtual sal_Int32 SAL_CALL close() = 0;

  //---------------------------------------------------------------------------
  /** gets the security ticket
   */
  virtual sal_Int32 SAL_CALL getSecurityTicket( ::rtl::ByteSequence& rTicket ) = 0;
  
  //---------------------------------------------------------------------------
  /** gets the user name
   */
  virtual ::rtl::OUString SAL_CALL getUser() = 0;

  //---------------------------------------------------------------------------
  /** gets the source address 
   */
  virtual ::rtl::OUString SAL_CALL getSource() = 0;

  //---------------------------------------------------------------------------
  /** gets the destination address 
   */
  virtual ::rtl::OUString SAL_CALL getDestination() = 0;
};


//---------------------------------------------------------------------------
/** interface to receive a connetion to a portal server
 */

class IPortalConnector: public ::vos::IReference
{
public:

  //---------------------------------------------------------------------------
  /** establishes a connection to a portal service

	  @param rUser        [in] : the user name for StarPortal
	  @param rPassword    [in] : the password used for logon
	  @param rServer      [in] : the server name as alias or IP/Port
	  @param rServiceName [in] : the service to connect to
	  @param rConnection  [out]: the connection object on success.

	  @returns E_None on success, an error value otherwise.
   */

  virtual RC SAL_CALL connectToService
  (
   const ::rtl::OUString& rUser,
   const ::rtl::OUString& rPassword,
   const ::rtl::OUString& rServer,
   const ::rtl::OUString& rServiceName,
   ::vos::ORef< IConnection >& rConnection
   ) = 0;
 
  //---------------------------------------------------------------------------
  /** establishes a connection to a portal service

	  @param rUser        [in] : the user name for StarPortal
	  @param rTicket      [in] : the security ticket used for logon
	  @param rServer      [in] : the server name as alias or IP/Port
	  @param rServiceName [in] : the service to connect to
	  @param rConnection  [out]: the connection object on success.

	  @returns E_None on success, an error value otherwise.
   */

  virtual RC SAL_CALL connectToService
  (
   const ::rtl::OUString& rUser,
   const ::rtl::ByteSequence& rTicket,
   const ::rtl::OUString& rServer,
   const ::rtl::OUString& rServiceName,
   ::vos::ORef< IConnection >& rConnection
   ) = 0;

};

//---------------------------------------------------------------------------
/** get the global connector object
 */

::vos::ORef< IPortalConnector > SAL_CALL getPortalConnector(::rtl::OUString &protocol);


//---------------------------------------------------------------------------
/** interface to receive a connetion to a portal client
 */

class IPortalAcceptor: public ::vos::IReference
{
public:

  //---------------------------------------------------------------------------
  /** accepts a connections from the portal daemon
   */
  virtual RC SAL_CALL acceptConnection( ::vos::ORef< IConnection >& rConnection ) = 0;

  //---------------------------------------------------------------------------
  /** stops any acceptConnection calls
   */
  virtual void SAL_CALL stopAccepting() = 0;
}; 


//---------------------------------------------------------------------------
/** get the global acceptor object
 */

::vos::ORef< IPortalAcceptor > SAL_CALL getPortalAcceptor(::rtl::OUString &protocol);

} // namespace daemons


//---------------------------------------------------------------------------
/** this is an ugly hack to load the new portalconnect lib dynamicly.
    will be removed as soon as the UNO XConnection supports non
    blocking read.
 */

extern "C" 
{
    /** get the global connector object
        <BR>
        if the pointer returned is not NULL, it represents a <type scope="daemons">IPortalConnector</type> object
        which is acquired once and has to be released once by the caller.
    */
	void* SAL_CALL _getPortalConnector();
}


#endif /* __cplusplus */
#endif /* _DAEMONS_PORTALCONNECT_HXX_ */


#endif






