/*************************************************************************
 *
 *  $RCSfile: chanpool.c,v $
 *
 *  $Revision: 1.1.1.1 $
 *
 *  last change: $Author: hr $ $Date: 2000/09/18 15:17:15 $
 *
 *  The Contents of this file are made available subject to the terms of
 *  either of the following licenses
 *
 *         - GNU Lesser General Public License Version 2.1
 *         - Sun Industry Standards Source License Version 1.1
 *
 *  Sun Microsystems Inc., October, 2000
 *
 *  GNU Lesser General Public License Version 2.1
 *  =============================================
 *  Copyright 2000 by Sun Microsystems, Inc.
 *  901 San Antonio Road, Palo Alto, CA 94303, USA
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *
 *  Sun Industry Standards Source License Version 1.1
 *  =================================================
 *  The contents of this file are subject to the Sun Industry Standards
 *  Source License Version 1.1 (the "License"); You may not use this file
 *  except in compliance with the License. You may obtain a copy of the
 *  License at http://www.openoffice.org/license.html.
 *
 *  Software provided under this License is provided on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 *  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
 *  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
 *  See the License for the specific provisions governing your rights and
 *  obligations concerning the Software.
 *
 *  The Initial Developer of the Original Code is: Sun Microsystems, Inc.
 *
 *  Copyright: 2000 by Sun Microsystems, Inc.
 *
 *  All Rights Reserved.
 *
 *  Contributor(s): _______________________________________
 *
 *
 ************************************************************************/

#include <osl/mutex.h>
#include <osl/conditn.h>

#include <string.h>

#include "chanpool.h"


typedef struct _PoolEntry
{
	sal_Bool				bPushed;
	oslChannelType			type;
	oslChannelResource		resource;
	rtl_uString*			ustrName;
	rtl_uString*			ustrUserIdent;
	rtl_uString*			ustrPassword;
	CheckAuthentication_F	pCheckFunc;
	void*					pCheckFuncData;
	oslCondition			readyCondition;
	oslChannelError			error;					
	struct _PoolEntry* 		pPrev;
	struct _PoolEntry* 		pNext;

} PoolEntry;


typedef struct 
{
	PoolEntry* 	pFirst;
	PoolEntry* 	pLast;
	oslMutex	mutex;
			
} ChannelPool;


/*****************************************************************************/
/* Static Module Functions */
/*****************************************************************************/

static void removeEntry( ChannelPool* pPool, PoolEntry* pEntry );

static PoolEntry* addEntry ( ChannelPool*           pPool, 
							 sal_Bool               bPushed,
							 rtl_uString*           ustrName,
							 rtl_uString*           ustrUserIdent,
							 rtl_uString*           ustrPassword,
							 CheckAuthentication_F  pFunc,
							 void*                  pFuncData );

static PoolEntry* searchEntry( ChannelPool*         pPool, 
							   sal_Bool             bPushed,
							   rtl_uString*         ustrName );



/*****************************************************************************/
/* createChannelPool_Impl */
/*****************************************************************************/

oslChannelPool createChannelPool_Impl()
{
	ChannelPool* pPool = (ChannelPool*) malloc(sizeof(ChannelPool));

	pPool->pFirst = NULL;
	pPool->pLast = NULL;
	pPool->mutex = osl_createMutex();

	return(pPool);
}	


/*****************************************************************************/
/* destroyChannelPool_Impl */
/*****************************************************************************/

void destroyChannelPool_Impl(oslChannelPool hPool)
{
	ChannelPool* pPool = (ChannelPool*) hPool;

	if (pPool != NULL)
	{
		osl_destroyMutex(pPool->mutex);
		free(pPool);
	}
}	


/*****************************************************************************/
/* channelPool_push_Impl */
/*****************************************************************************/

oslChannelError channelPool_push_Impl
(
	oslChannelPool 			hPool, 	
	oslChannelType 			type,
	oslChannelResource 		resource,
	rtl_uString* 		    ustrName,
	rtl_uString* 		    ustrUserIdent,
	rtl_uString* 		    ustrPassword,
	const TimeValue* 		pTimeout
)
{
	oslChannelError error = osl_Channel_E_Unknown;
	ChannelPool* pPool = (ChannelPool*) hPool;

	if (pPool != NULL)
	{
		PoolEntry* pEntry = NULL;
		
		osl_acquireMutex(pPool->mutex);

		/* is any thread waiting for this channel ? */
		pEntry = searchEntry(pPool, sal_False, ustrName);

		if (pEntry != NULL)
		{
			sal_Bool bAuthOk = sal_True;

			if (pEntry->pCheckFunc)
			{
				/* check authentication */
				bAuthOk = pEntry->pCheckFunc(pEntry->pCheckFuncData, ustrUserIdent, ustrPassword);	
			}

			if (bAuthOk)
			{
				/* yes: give the resource to him and post ready condition to acceptor! */
				pEntry->resource = resource;
				pEntry->type = type;			
				pEntry->error = osl_Channel_E_None;			
				osl_setCondition(pEntry->readyCondition);

				error = osl_Channel_E_None;			
			}
			else
				error = osl_Channel_E_AuthenticationFailed;			
		}
		else
		{
			if ((pTimeout == NULL) || (pTimeout->Seconds > 0)|| (pTimeout->Nanosec > 0))
			{
				/* no and timeout > 0: add to pool and wait for any thread who want this channel! */
				pEntry = addEntry(
							pPool, 
							sal_True, 
							ustrName, 
							ustrUserIdent, 
							ustrPassword,
							NULL,
							NULL);

				pEntry->type = type;
				pEntry->resource = resource;

				osl_releaseMutex(pPool->mutex);
				osl_waitCondition(pEntry->readyCondition, pTimeout);
				osl_acquireMutex(pPool->mutex);
				
				error = pEntry->error;

				removeEntry(pPool, pEntry);
			}
			else
			{
				/* no and timeout == 0: timeout reached! */
				error = osl_Channel_E_Timeout;
			}
		}

		osl_releaseMutex(pPool->mutex);
	}
	
	return(error);
}	


/*****************************************************************************/
/* channelPool_pop_Impl */
/*****************************************************************************/

oslChannelError channelPool_pop_Impl
(
	oslChannelPool 			hPool, 	
	oslChannelType* 		pType,
	oslChannelResource*		pResource,
	rtl_uString*     		ustrName,
   	CheckAuthentication_F	checkAuthFunc,
	void* 					authFuncData,
	const TimeValue* 		pTimeout
)
{
	oslChannelError error = osl_Channel_E_Unknown;
	ChannelPool* pPool = (ChannelPool*) hPool;

	if (pPool != NULL)
	{
		PoolEntry* pEntry = NULL;
		
		osl_acquireMutex(pPool->mutex);

		/* is any thread providing for this channel ? */
		pEntry = searchEntry(pPool, sal_True, ustrName);

		if (pEntry != NULL)
		{
			sal_Bool bAuthOk = sal_True;

			if (checkAuthFunc)
			{
				/* check authentication */
				bAuthOk = checkAuthFunc(authFuncData, pEntry->ustrUserIdent, pEntry->ustrPassword);
			}


			if (bAuthOk)
			{
				/* yes: get the resource and post ready condition to provider! */
				*pResource = pEntry->resource;
				*pType = pEntry->type;			
				pEntry->error = osl_Channel_E_None;			
				osl_setCondition(pEntry->readyCondition);

				error = osl_Channel_E_None;			
			}
			else
				error = osl_Channel_E_AuthenticationFailed;			
		}
		else
		{
			if ((pTimeout == NULL) || (pTimeout->Seconds > 0)|| (pTimeout->Nanosec > 0))
			{
				/* no and timeout > 0: add to pool and wait for any thread who want this channel! */
				pEntry = addEntry(
							pPool, 
							sal_False, 
							ustrName, 
							NULL,
							NULL,
							checkAuthFunc, 
							authFuncData);

				osl_releaseMutex(pPool->mutex);
				osl_waitCondition(pEntry->readyCondition, pTimeout);
				osl_acquireMutex(pPool->mutex);
				
				error = pEntry->error;
				
				if (error == osl_Channel_E_None)
				{
					*pResource = pEntry->resource;
					*pType = pEntry->type;			
				}

				removeEntry(pPool, pEntry);
			}
			else
			{
				/* no and timeout == 0: timeout reached! */
				error = osl_Channel_E_Timeout;
			}
		}

		osl_releaseMutex(pPool->mutex);
	}
	
	return(error);
}	



/*****************************************************************************/
/* addEntry */
/*****************************************************************************/

static PoolEntry* searchEntry
(
	ChannelPool* 	pPool, 
	sal_Bool 		bPushed,
	rtl_uString*    ustrName
)
{
	PoolEntry** ppEntry = &(pPool->pFirst);
	sal_Bool bFound = sal_False;

	while (((*ppEntry) != NULL) && !bFound)
	{
		if (
			((*ppEntry)->bPushed == bPushed) &&
			(rtl_ustr_compare((*ppEntry)->ustrName->buffer, ustrName->buffer) == 0)
		   )
		{
			bFound = sal_True;
		}
		else
			ppEntry = &((*ppEntry)->pNext);
	}

	return(*ppEntry);
}	


/*****************************************************************************/
/* addEntry */
/*****************************************************************************/

static PoolEntry* addEntry
(
	ChannelPool* 			pPool, 
	sal_Bool 				bPushed,
	rtl_uString*    		ustrName,
	rtl_uString*    		ustrUserIdent,
	rtl_uString* 	    	ustrPassword,
	CheckAuthentication_F	pFunc,
	void*					pFuncData
)
{
	PoolEntry* pEntry = (PoolEntry*) malloc(sizeof(PoolEntry));
    memset(pEntry,0,sizeof(PoolEntry));
    
	pEntry->bPushed = bPushed;
	pEntry->readyCondition = osl_createCondition();
	pEntry->error = osl_Channel_E_Timeout;
	pEntry->pCheckFunc = pFunc;
	pEntry->pCheckFuncData = pFuncData;

    rtl_uString_acquire(ustrName);    
    pEntry->ustrName=ustrName;

    rtl_uString_acquire(ustrName);
	pEntry->ustrUserIdent = ustrUserIdent;

    rtl_uString_acquire(ustrName); 
	pEntry->ustrPassword = ustrPassword;

	if (pPool->pFirst == NULL)
	{
		pPool->pFirst = pEntry;
		pPool->pLast = pEntry;
		pEntry->pPrev = NULL;
		pEntry->pNext = NULL;
	}
	else
	{
		pEntry->pPrev = pPool->pLast;
		pEntry->pPrev->pNext = pEntry;
		pEntry->pNext = NULL;
		pPool->pLast = pEntry;
	}

	return(pEntry);
}	


/*****************************************************************************/
/* removeEntry */
/*****************************************************************************/

static void removeEntry
(
	ChannelPool* 	pPool, 
	PoolEntry* 		pEntry
)
{
	if (pEntry->pPrev == NULL)
		pPool->pFirst = pEntry->pNext;
	else
		pEntry->pPrev->pNext = pEntry->pNext;

	if (pEntry->pNext == NULL)
		pPool->pLast = pEntry->pPrev;
	else
		pEntry->pNext->pPrev = pEntry->pPrev;

	osl_destroyCondition(pEntry->readyCondition);
	free(pEntry);
}	


