/*************************************************************************
 *
 *  $RCSfile: socket.c,v $
 *
 *  $Revision: 1.1.1.1 $
 *
 *  last change: $Author: hr $ $Date: 2000/09/18 15:17:23 $
 *
 *  The Contents of this file are made available subject to the terms of
 *  either of the following licenses
 *
 *         - GNU Lesser General Public License Version 2.1
 *         - Sun Industry Standards Source License Version 1.1
 *
 *  Sun Microsystems Inc., October, 2000
 *
 *  GNU Lesser General Public License Version 2.1
 *  =============================================
 *  Copyright 2000 by Sun Microsystems, Inc.
 *  901 San Antonio Road, Palo Alto, CA 94303, USA
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *
 *  Sun Industry Standards Source License Version 1.1
 *  =================================================
 *  The contents of this file are subject to the Sun Industry Standards
 *  Source License Version 1.1 (the "License"); You may not use this file
 *  except in compliance with the License. You may obtain a copy of the
 *  License at http://www.openoffice.org/license.html.
 *
 *  Software provided under this License is provided on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 *  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
 *  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
 *  See the License for the specific provisions governing your rights and
 *  obligations concerning the Software.
 *
 *  The Initial Developer of the Original Code is: Sun Microsystems, Inc.
 *
 *  Copyright: 2000 by Sun Microsystems, Inc.
 *
 *  All Rights Reserved.
 *
 *  Contributor(s): _______________________________________
 *
 *
 ************************************************************************/

#include "system.h"

#include <osl/socket.h>
#include <osl/diagnose.h>

#include "sockimpl.h"

/* defines for shutdown */
#ifdef GCC
#	define SD_RECEIVE 0
#	define SD_SEND 1
#	define SD_BOTH 2
#endif

/*
	oslSocketAddr is a pointer to a Berkeley struct sockaddr.
	I refrained from using sockaddr_in because of possible further
	extensions of this socket-interface (IP-NG?). 
	The intention was to hide all Berkeley data-structures from
	direct access past the osl-interface.

	The current implementation is internet (IP) centered. All
	the constructor-functions (osl_create...) take parameters
	that will probably make sense only in the IP-environment
	(e.g. because of using the dotted-Addr-format).
	
	If the interface will be extended to host other protocol-
	families, I expect no externally visible changes in the
	existing functions. You'll probably need only new
	constructor-functions who take the different Addr
	formats into consideration (maybe a long dotted Addr
	or whatever).
*/

/*
  _Note_ that I rely on the fact that oslSocketAddr and struct sockaddr
  are the same! I don't like it very much but see no other easy way to
  conceal the struct sockaddr from the eyes of the user.
*/

#define OSL_INVALID_SOCKET		INVALID_SOCKET			/* WIN32 */
#define OSL_SOCKET_ERROR		SOCKET_ERROR			/* WIN32 */

/*****************************************************************************/
/* enum oslAddrFamily */
/*****************************************************************************/

/* map */
static DWORD FamilyMap[]= {
	AF_INET,					/* osl_Socket_FamilyInet */
	AF_IPX,						/* osl_Socket_FamilyIpx */
	0							/* osl_Socket_FamilyInvalid */
};

/* reverse map */
static oslAddrFamily osl_AddrFamilyFromNative(DWORD nativeType)
{
	oslAddrFamily i= 0; 
	while(i != osl_Socket_FamilyInvalid)
	{
		if(FamilyMap[i] == nativeType)
			return i;
		i++;
	}
	return i;
}

/* macros */
#define FAMILY_FROM_NATIVE(y) osl_AddrFamilyFromNative(y)
#define FAMILY_TO_NATIVE(x)	(short)FamilyMap[x]

/*****************************************************************************/
/* enum oslProtocol */
/*****************************************************************************/

/* map */
static DWORD ProtocolMap[]= {
	0,							/* osl_Socket_FamilyInet */
	NSPROTO_IPX,				/* osl_Socket_FamilyIpx */
	NSPROTO_SPX,				/* osl_Socket_ProtocolSpx */
	NSPROTO_SPXII,				/* osl_Socket_ProtocolSpx_ii */
	0							/* osl_Socket_ProtocolInvalid */
};

/* reverse map */
static oslProtocol osl_ProtocolFromNative(DWORD nativeType)
{
	oslProtocol i= 0; 
	while(i != osl_Socket_ProtocolInvalid)
	{
		if(ProtocolMap[i] == nativeType)
			return i;
		i++;
	}
	return i;
}

/* macros */
#define PROTOCOL_FROM_NATIVE(y) osl_ProtocolFromNative(y)
#define PROTOCOL_TO_NATIVE(x)	ProtocolMap[x]

/*****************************************************************************/
/* enum oslSocketType */
/*****************************************************************************/

/* map */
static DWORD TypeMap[]= {
	SOCK_STREAM,				/* osl_Socket_TypeStream */
	SOCK_DGRAM,					/* osl_Socket_TypeDgram  */
	SOCK_RAW,					/* osl_Socket_TypeRaw */
	SOCK_RDM,					/* osl_Socket_TypeRdm */
	SOCK_SEQPACKET,				/* osl_Socket_TypeSeqPacket */
	0							/* osl_Socket_TypeInvalid */
};

/* reverse map */
static oslSocketType osl_SocketTypeFromNative(DWORD nativeType)
{
	oslSocketType i= 0; 
	while(i != osl_Socket_TypeInvalid)
	{
		if(TypeMap[i] == nativeType)
			return i;
		i++;
	}
	return i;
}

/* macros */
#define TYPE_TO_NATIVE(x)		TypeMap[x]
#define TYPE_FROM_NATIVE(y)		osl_SocketTypeFromNative(y)

/*****************************************************************************/
/* enum oslSocketOption */
/*****************************************************************************/

/* map */
static DWORD OptionMap[]= {
	SO_DEBUG,					/* osl_Socket_OptionDebug */
	SO_ACCEPTCONN,				/* osl_Socket_OptionAcceptConn */
	SO_REUSEADDR,				/* osl_Socket_OptionReuseAddr */
	SO_KEEPALIVE,				/* osl_Socket_OptionKeepAlive */
	SO_DONTROUTE,				/* osl_Socket_OptionDontRoute */
	SO_BROADCAST,				/* osl_Socket_OptionBroadcast */
	SO_USELOOPBACK,				/* osl_Socket_OptionUseLoopback */
	SO_LINGER,					/* osl_Socket_OptionLinger */
	SO_OOBINLINE,				/* osl_Socket_OptionOOBinLine */
	SO_SNDBUF,					/* osl_Socket_OptionSndBuf */
	SO_RCVBUF,					/* osl_Socket_OptionRcvBuf */
	SO_SNDLOWAT,				/* osl_Socket_OptionSndLowat */
	SO_RCVLOWAT,				/* osl_Socket_OptionRcvLowat */
	SO_SNDTIMEO,				/* osl_Socket_OptionSndTimeo */
	SO_RCVTIMEO,				/* osl_Socket_OptionRcvTimeo */
	SO_ERROR,					/* osl_Socket_OptionError */
	SO_TYPE,					/* osl_Socket_OptionType */
	TCP_NODELAY,				/* osl_Socket_OptionTcpNoDelay */
	0							/* osl_Socket_OptionInvalid */
};	

/* reverse map */
static oslSocketOption osl_SocketOptionFromNative(DWORD nativeType)
{
	oslSocketOption i= 0; 
	while(i != osl_Socket_OptionInvalid)
	{
		if(OptionMap[i] == nativeType)
			return i;
		i++;
	}
	return i;
}

/* macros */
#define OPTION_TO_NATIVE(x)		OptionMap[x]
#define OPTION_FROM_NATIVE(y)	osl_SocketOptionFromNative(y)

/*****************************************************************************/
/* enum oslSocketOptionLevel */
/*****************************************************************************/

static DWORD OptionLevelMap[]= {
	SOL_SOCKET,					/* osl_Socket_LevelSocket */
	IPPROTO_TCP,				/* osl_Socket_LevelTcp */
	0							/* osl_invalid_SocketLevel */
};

/* reverse map */
static oslSocketOptionLevel osl_SocketOptionLevelFromNative(DWORD nativeType)
{
	oslSocketOptionLevel i= 0; 
	while(i != osl_Socket_LevelInvalid)
	{
		if(OptionLevelMap[i] == nativeType)
			return i;
		i++;
	}
	return i;
}

/* macros */
#define OPTION_LEVEL_TO_NATIVE(x)		OptionLevelMap[x]
#define OPTION_LEVEL_FROM_NATIVE(y)		osl_SocketOptionLevelFromNative(y)

/*****************************************************************************/
/* enum oslSocketMsgFlag */
/*****************************************************************************/

static DWORD SocketMsgFlagMap[]= {
	0,							/* osl_Socket_MsgNormal */
	MSG_OOB,					/* osl_Socket_MsgOOB */
	MSG_PEEK,					/* osl_Socket_MsgPeek */
	MSG_DONTROUTE,				/* osl_Socket_MsgDontRoute */
	MSG_MAXIOVLEN				/* osl_Socket_MsgMaxIOVLen */
};

/* reverse map */
static oslSocketMsgFlag osl_SocketMsgFlagFromNative(DWORD nativeType)
{
	oslSocketMsgFlag i= 0; 

	while(i != osl_Socket_MsgInvalid)
	{
		if(SocketMsgFlagMap[i] == nativeType)
			return i;
	
		i++;
	}
	return i;
}

/* macros */
#define MSG_FLAG_TO_NATIVE(x)		SocketMsgFlagMap[x]
#define MSG_FLAG_FROM_NATIVE(y)		osl_SocketMsgFlagFromNative(y)

/*****************************************************************************/
/* enum oslSocketDirection */
/*****************************************************************************/

static DWORD SocketDirection[]= {
	SD_RECEIVE,					/* osl_Socket_DirRead */
	SD_SEND,					/* osl_Socket_DirWrite */
	SD_BOTH						/* osl_Socket_DirReadwrite */
};

/* reverse map */
static oslSocketDirection osl_SocketDirectionFromNative(DWORD nativeType)
{
	oslSocketDirection i= 0; 

	while(i != osl_Socket_DirInvalid)
	{
		if(SocketDirection[i] == nativeType)
			return i;
		i++;
	}
	
	return i;
}

/* macros */
#define DIRECTION_TO_NATIVE(x)		SocketDirection[x]
#define DIRECTION_FROM_NATIVE(y)	osl_SocketDirectionFromNative(y)

/*****************************************************************************/
/* enum oslSocketError */
/*****************************************************************************/

static int SocketError[]= {

	0,					/* no error */
	WSAENOTSOCK,			/* Socket operation on non-socket */
	WSAEDESTADDRREQ,		/* Destination address required */
	WSAEMSGSIZE,			/* Message too long */
	WSAEPROTOTYPE,			/* Protocol wrong type for socket */
	WSAENOPROTOOPT,		/* Protocol not available */
	WSAEPROTONOSUPPORT,	/* Protocol not supported */
	WSAESOCKTNOSUPPORT,	/* Socket type not supported */
	WSAEOPNOTSUPP,			/* Operation not supported on socket */
	WSAEPFNOSUPPORT,		/* Protocol family not supported */
	WSAEAFNOSUPPORT,		/* Address family not supported by */
							/* protocol family */
	WSAEADDRINUSE,			/* Address already in use */
	WSAEADDRNOTAVAIL,		/* Can't assign requested address */
	WSAENETDOWN,			/* Network is down */
	WSAENETUNREACH,		/* Network is unreachable */
	WSAENETRESET,			/* Network dropped connection because */
							/* of reset */
	WSAECONNABORTED,		/* Software caused connection abort */
	WSAECONNRESET,			/* Connection reset by peer */
	WSAENOBUFS,			/* No buffer space available */
	WSAEISCONN,			/* Socket is already connected */
	WSAENOTCONN,			/* Socket is not connected */
	WSAESHUTDOWN,			/* Can't send after socket shutdown */
	WSAETOOMANYREFS,		/* Too many references: can't splice */
	WSAETIMEDOUT,			/* Connection timed out */
	WSAECONNREFUSED,		/* Connection refused */
	WSAEHOSTDOWN,			/* Host is down */
	WSAEHOSTUNREACH,		/* No route to host */
	WSAEWOULDBLOCK,		/* call would block on non-blocking socket */
	WSAEALREADY,			/* operation already in progress */
	WSAEINPROGRESS		/* operation now in progress */
};

/* reverse map */
static oslSocketError osl_SocketErrorFromNative(int nativeType)
{
	oslSocketError i= 0; 

	while(i != osl_Socket_E_InvalidError)
	{
		if(SocketError[i] == nativeType)
			return i;
	
		i++;
	}
	return i;
}

/* macros */
#define ERROR_TO_NATIVE(x)		SocketError[x]
#define ERROR_FROM_NATIVE(y)	osl_SocketErrorFromNative(y)

/*****************************************************************************/
/* oslSocketDialupImpl */
/*****************************************************************************/
static oslSocketDialupImpl *pDialupImpl = NULL;

#if 0  /* INTERNAL DEBUG ONLY */
BOOL WINAPI __osl_autodial_Impl (DWORD dwFlags, DWORD dwReserved)
{
	return 0;
}

BOOL WINAPI __osl_autodialHangup_Impl (DWORD dwReserved)
{
	return 1;
}

BOOL WINAPI __osl_getConnectedState_Impl (LPDWORD lpdwFlags, DWORD dwReserved)
{
	if (lpdwFlags)
		*lpdwFlags = 0;
	return 0;
}
#endif /* INTERNAL DEBUG ONLY */

/*
 * __osl_createSocketDialupImpl.
 */
static oslSocketDialupImpl* __osl_createSocketDialupImpl (void)
{
	oslSocketDialupImpl *pImpl;
	pImpl = (oslSocketDialupImpl*)calloc(1, sizeof (oslSocketDialupImpl));

	InitializeCriticalSection (&pImpl->m_hMutex);

	return (pImpl);
}

/*
 * __osl_initSocketDialupImpl.
 */
static void __osl_initSocketDialupImpl (oslSocketDialupImpl *pImpl)
{
	if (pImpl)
	{
		HINSTANCE hModule;

		EnterCriticalSection (&pImpl->m_hMutex);

		hModule = LoadLibrary (INTERNET_MODULE_NAME);
		if (!(hModule <= (HINSTANCE)HINSTANCE_ERROR))
		{
			pImpl->m_pfnAttemptConnect = (INTERNETATTEMPTCONNECT)
				(GetProcAddress (hModule, "InternetAttemptConnect"));
			pImpl->m_pfnAutodial = (INTERNETAUTODIAL)
				(GetProcAddress (hModule, "InternetAutodial"));
			pImpl->m_pfnAutodialHangup = (INTERNETAUTODIALHANGUP)
				(GetProcAddress (hModule, "InternetAutodialHangup"));
			pImpl->m_pfnGetConnectedState = (INTERNETGETCONNECTEDSTATE)
				(GetProcAddress (hModule, "InternetGetConnectedState"));
			pImpl->m_hModule = hModule;
		}

		LeaveCriticalSection (&pImpl->m_hMutex);
	}
}

/*
 * __osl_destroySocketDialupImpl.
 */
static void __osl_destroySocketDialupImpl (oslSocketDialupImpl *pImpl)
{
	if (pImpl)
	{
		EnterCriticalSection (&pImpl->m_hMutex);

		if (pImpl->m_dwFlags & INTERNET_CONNECTION_HANGUP)
		{
			if (pImpl->m_pfnAutodialHangup)
			{
				(pImpl->m_pfnAutodialHangup)(0);
				pImpl->m_dwFlags &= ~INTERNET_CONNECTION_HANGUP;
			}
		}

		if (pImpl->m_hModule)
			FreeLibrary (pImpl->m_hModule);

		LeaveCriticalSection (&pImpl->m_hMutex);
		DeleteCriticalSection (&pImpl->m_hMutex);

		free (pImpl);
	}
}

/*
 * __osl_querySocketDialupImpl.
 */
static sal_Bool __osl_querySocketDialupImpl (void)
{
	sal_Bool result;

	if (pDialupImpl == NULL)
	{
		pDialupImpl = __osl_createSocketDialupImpl();
		__osl_initSocketDialupImpl (pDialupImpl);
	}

	EnterCriticalSection (&pDialupImpl->m_hMutex);

	result = sal_True;
	if (pDialupImpl->m_pfnGetConnectedState)
	{
		DWORD dwFlags = 0;

		result = (pDialupImpl->m_pfnGetConnectedState)(&dwFlags, 0);
		pDialupImpl->m_dwFlags |= dwFlags;
	}

	LeaveCriticalSection (&pDialupImpl->m_hMutex);
	return (result);
}

/*
 * __osl_attemptSocketDialupImpl.
 */
static sal_Bool __osl_attemptSocketDialupImpl (void)
{
	sal_Bool result;

	if (pDialupImpl == NULL)
	{
		pDialupImpl = __osl_createSocketDialupImpl();
		__osl_initSocketDialupImpl (pDialupImpl);
	}

	EnterCriticalSection (&pDialupImpl->m_hMutex);

	result = __osl_querySocketDialupImpl();
	if (!result)
	{
		result = sal_True;
		if (pDialupImpl->m_pfnAutodial)
		{
			result = (pDialupImpl->m_pfnAutodial)(0, 0);
			if (result)
				pDialupImpl->m_dwFlags |= INTERNET_CONNECTION_HANGUP;
			else
				WSASetLastError (WSAENETDOWN);
		}
	}

	LeaveCriticalSection (&pDialupImpl->m_hMutex);
	return (result);
}

/*****************************************************************************/
/* oslSocketImpl */
/*****************************************************************************/
static sal_uInt32 nSocketImpl = 0;

/*
 * __osl_createSocketImpl.
 */
oslSocketImpl* __osl_createSocketImpl(SOCKET Socket)
{
	oslSocketImpl* pSockImpl;

	pSockImpl = (oslSocketImpl*)calloc(1, sizeof(oslSocketImpl));
	pSockImpl->m_Socket = Socket;

	nSocketImpl++;
	
	return (pSockImpl);
}

/*
 * __osl_destroySocketImpl.
 */
void __osl_destroySocketImpl(oslSocketImpl *pImpl)
{
	if (pImpl)
	{
		if (--nSocketImpl == 0)
		{
			__osl_destroySocketDialupImpl (pDialupImpl);
			pDialupImpl = NULL;
		}
		free (pImpl);
	}
}

/*****************************************************************************/
/* osl_createEmptySocketAddr */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_createEmptySocketAddr(oslAddrFamily Family)
{
	struct sockaddr* pAddr;

	pAddr= malloc(sizeof(struct sockaddr));

	/* is it an internet-Addr? */
	if (Family == osl_Socket_FamilyInet)
	{
		struct sockaddr_in* pInetAddr= (struct sockaddr_in*)pAddr;

		pInetAddr->sin_family = FAMILY_TO_NATIVE(osl_Socket_FamilyInet);
		pInetAddr->sin_addr.s_addr = htonl(INADDR_ANY);
		pInetAddr->sin_port = 0;
	} 
	else 
	{
		memset(pAddr, 0, sizeof(struct sockaddr));
		pAddr->sa_family = FAMILY_TO_NATIVE(Family);
	}

	return (oslSocketAddr)pAddr;
}

/*****************************************************************************/
/* osl_copySocketAddr */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_copySocketAddr(oslSocketAddr Addr)
{
	struct sockaddr* pAddr= NULL;

	if (Addr)
	{
		pAddr= malloc(sizeof(struct sockaddr));
		OSL_ASSERT(pAddr);

		if (pAddr)
			memcpy(pAddr, Addr, sizeof(struct sockaddr));
	}

	return pAddr;
}

/*****************************************************************************/
/* osl_isEqualSocketAddr */
/*****************************************************************************/
sal_Bool SAL_CALL osl_isEqualSocketAddr(oslSocketAddr Addr1, oslSocketAddr Addr2)
{
	struct sockaddr* pAddr1= (struct sockaddr*)Addr1;
	struct sockaddr* pAddr2= (struct sockaddr*)Addr2;

	OSL_ASSERT(pAddr1);
	OSL_ASSERT(pAddr2);

	if (pAddr1->sa_family == pAddr2->sa_family)
	{
		switch (pAddr1->sa_family)
		{
			case AF_INET:
			{
				struct sockaddr_in* pInetAddr1= (struct sockaddr_in*)pAddr1;
				struct sockaddr_in* pInetAddr2= (struct sockaddr_in*)pAddr2;

				if ((pInetAddr1->sin_family == pInetAddr2->sin_family) &&
					(pInetAddr1->sin_addr.s_addr == pInetAddr2->sin_addr.s_addr) &&
					(pInetAddr1->sin_port == pInetAddr2->sin_port))
					return (sal_True);
			}

			case AF_IPX:
			{
				struct sockaddr_ipx* pIpxAddr1= (struct sockaddr_ipx*)pAddr1;
				struct sockaddr_ipx* pIpxAddr2= (struct sockaddr_ipx*)pAddr2;

				if ((pIpxAddr1->sa_family == pIpxAddr2->sa_family) &&
					(memcmp(&pIpxAddr1->sa_netnum, &pIpxAddr2->sa_netnum, sizeof(pIpxAddr1->sa_netnum)) == 0) &&
					(memcmp(&pIpxAddr1->sa_nodenum, &pIpxAddr2->sa_nodenum, sizeof(pIpxAddr1->sa_nodenum)) == 0) &&
					(pIpxAddr1->sa_socket == pIpxAddr2->sa_socket))
					return (sal_True);
			}

			default:
			{
				return (memcmp(pAddr1, Addr2, sizeof(struct sockaddr)) == 0);
			}
		}
	}

	return (sal_False);
}

/*****************************************************************************/
/* osl_createInetBroadcastAddr */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_createInetBroadcastAddr (
	rtl_uString *strDottedAddr,
	sal_Int32    Port)
{
	struct sockaddr_in *pAddr = NULL;
	sal_uInt32          nAddr = OSL_INADDR_NONE;

	if (strDottedAddr && strDottedAddr->length)
	{
		/* Dotted host address for limited broadcast */
		rtl_String *pDottedAddr = NULL;

		rtl_uString2String (
			&pDottedAddr, strDottedAddr->buffer, strDottedAddr->length,
			RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS);

		nAddr = inet_addr (pDottedAddr->buffer);
		rtl_string_release (pDottedAddr);
	}

	if (nAddr != OSL_INADDR_NONE)
	{
		/* Limited broadcast */
		nAddr = ntohl(nAddr);
		if (IN_CLASSA(nAddr))
		{
			nAddr &= IN_CLASSA_NET;
			nAddr |= IN_CLASSA_HOST;
		}
		else if (IN_CLASSB(nAddr))
		{
			nAddr &= IN_CLASSB_NET;
			nAddr |= IN_CLASSB_HOST;
		}
		else if (IN_CLASSC(nAddr))
		{
			nAddr &= IN_CLASSC_NET;
			nAddr |= IN_CLASSC_HOST;
		}
		else
		{
			/* No broadcast in class D */
			return ((oslSocketAddr)NULL);
		}
		nAddr = htonl(nAddr);
	}

	pAddr = malloc (sizeof(struct sockaddr_in));
	if (pAddr)
	{
		pAddr->sin_family      = FAMILY_TO_NATIVE(osl_Socket_FamilyInet);
		pAddr->sin_port        = htons((sal_uInt16)(Port & 0xffff));
		pAddr->sin_addr.s_addr = nAddr; /* Already in network order */
	}
	return ((oslSocketAddr)pAddr);
}

/*****************************************************************************/
/* osl_createInetSocketAddr */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_createInetSocketAddr (
	rtl_uString *strDottedAddr, 
	sal_Int32    Port)
{
	DWORD Addr;
	rtl_String	*pDottedAddr=NULL;
	
	rtl_uString2String(
		&pDottedAddr, strDottedAddr->buffer, strDottedAddr->length,
		RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS);

	Addr= inet_addr (pDottedAddr->buffer);
	rtl_string_release (pDottedAddr);

	if(Addr != -1)
	{
		/* valid dotted addr */
		struct sockaddr_in* pInetAddr;

		pInetAddr= malloc(sizeof(struct sockaddr_in));
		OSL_ASSERT(pInetAddr);

		if (pInetAddr)
		{
			pInetAddr->sin_family      = FAMILY_TO_NATIVE(osl_Socket_FamilyInet);
			pInetAddr->sin_port        = htons((sal_uInt16)Port);
			pInetAddr->sin_addr.s_addr = Addr; /* already in right order */
		}

		return (oslSocketAddr)pInetAddr;
	}

	return (oslSocketAddr)0;
}

/*****************************************************************************/
/* osl_createIpxSocketAddr */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_createIpxSocketAddr (rtl_uString * strNetNumber,
									  			rtl_uString * strNodeNumber,
												sal_uInt32 SocketNumber)
{
	struct sockaddr_ipx* pIpxAddr;
	
	pIpxAddr= malloc(sizeof(struct sockaddr));
	OSL_ASSERT(pIpxAddr);

	if (pIpxAddr)
	{
		rtl_String	*NetNumber=NULL;
		rtl_String	*NodeNumber=NULL;
		pIpxAddr->sa_family= FAMILY_TO_NATIVE(osl_Socket_FamilyIpx);

		rtl_uString2String(
			&NetNumber,
			rtl_uString_getStr(strNetNumber),
			rtl_uString_getLength(strNetNumber),
			RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS );
		memcpy(&pIpxAddr->sa_netnum,
			   rtl_string_getStr( NetNumber),
			   sizeof(pIpxAddr->sa_netnum));
		
		rtl_uString2String(
			&NodeNumber,
			rtl_uString_getStr(strNodeNumber),
			rtl_uString_getLength(strNodeNumber),
			RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS );
		memcpy(&pIpxAddr->sa_nodenum,
			   rtl_string_getStr( NodeNumber),
			   sizeof(pIpxAddr->sa_nodenum));

		pIpxAddr->sa_socket= (sal_uInt16)SocketNumber;

		rtl_string_release(NetNumber);
		rtl_string_release(NodeNumber);
	}

	return (oslSocketAddr)pIpxAddr;
}

/*****************************************************************************/
/* oslHostAddr */
/*****************************************************************************/
typedef struct oslHostAddrImpl_t {
	rtl_uString     *pHostName;
	struct sockaddr *pSockAddr;
} oslHostAddrImpl;

static oslHostAddr __osl_hostentToHostAddr (const struct hostent *he)
{
	oslHostAddrImpl *pAddr= NULL;
	struct sockaddr *sa= NULL;
	rtl_uString     *cn= NULL;
#ifdef _DEBUG
	const sal_Char fct[] = "_osl_hostentToHostAddr()";
#endif /* _DEBUG */

	if ((he == NULL) || (he->h_name == NULL) || (he->h_addr_list[0] == NULL))
		return ((oslHostAddr)NULL);

	rtl_string2UString(
		&cn, he->h_name, strlen(he->h_name),
		RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS);

	sa= (struct sockaddr  *)malloc (sizeof (struct sockaddr));
	OSL_ASSERT(sa);
	if (sa == NULL)
	{
		rtl_uString_release(cn);
		return ((oslHostAddr)NULL);
	}
	memset (sa, 0, sizeof (struct sockaddr));

	sa->sa_family= he->h_addrtype;
	if (sa->sa_family == FAMILY_TO_NATIVE(osl_Socket_FamilyInet))
	{
		struct sockaddr_in *sin= (struct sockaddr_in *)sa;
		memcpy (
			&(sin->sin_addr.s_addr),
			he->h_addr_list[0],
			he->h_length); 
	}
	else if (sa->sa_family == FAMILY_TO_NATIVE(osl_Socket_FamilyIpx))
	{
		struct sockaddr_ipx *spx= (struct sockaddr_ipx *)sa;

		/* not quite sure what ipx/spx defines as a host-address:  */
		/* just the hostnumber of the combination of net- and hostnumber */
		/* I guess the later, but am prepared for anything */

		switch (he->h_length)
		{
			case sizeof(spx->sa_netnum) + sizeof(spx->sa_nodenum):
			case sizeof(spx->sa_netnum):
				memcpy (
					&(spx->sa_netnum),
					he->h_addr_list[0],
					he->h_length);
				break;

			case sizeof(spx->sa_nodenum):
				memcpy (
					&(spx->sa_nodenum),
					he->h_addr_list[0],
					he->h_length);
				break;

			default:
#ifdef DEBUG
				OSL_TRACE("%s: unknown IPX/SPX address format.\n", fct);
#endif
				OSL_ASSERT(sal_False);

				free (sa);
				rtl_uString_release(cn);
				return ((oslHostAddr)NULL);
		}
	}
	else
	{
		/* unknown address family */
		/* future extensions for new families might be implemented here */

#ifdef DEBUG
		OSL_TRACE("%s: unknown address family.\n", fct);
#endif
		OSL_ASSERT(sal_False);

		free (sa);
		rtl_uString_release(cn);		
		return ((oslHostAddr)NULL);
	}

	pAddr= (oslHostAddrImpl *)malloc (sizeof (oslHostAddrImpl));
	OSL_ASSERT(pAddr);
	if (pAddr == NULL)
	{
		free (sa);
		rtl_uString_release(cn);
		return ((oslHostAddr)NULL);
	}

	pAddr->pHostName= cn;
	pAddr->pSockAddr= sa;

	return ((oslHostAddr)pAddr);
}

/*****************************************************************************/
/* osl_createHostAddr */
/*****************************************************************************/
oslHostAddr SAL_CALL osl_createHostAddr (
	rtl_uString         *strHostname,
	const oslSocketAddr  Addr)
{
	oslHostAddrImpl *pAddr;
	struct sockaddr *sa;
	rtl_uString     *cn= NULL;

	OSL_ASSERT(strHostname && strHostname->length && Addr);
	if ((strHostname == NULL)  || (strHostname->length == 0) || (Addr == NULL))
		return ((oslHostAddr)NULL);

	rtl_uString_newFromString( &cn, strHostname);

	sa= (struct sockaddr *)osl_copySocketAddr (Addr);
	if (sa == NULL)
	{
		rtl_uString_release(cn);
		return ((oslHostAddr)NULL);
	}

	pAddr= (oslHostAddrImpl *)malloc (sizeof (oslHostAddrImpl));
	OSL_ASSERT(pAddr);
	if (pAddr == NULL)
	{
		rtl_uString_release(cn);		
		free (sa);
		return ((oslHostAddr)NULL);
	}

	pAddr->pHostName= cn;
	pAddr->pSockAddr= sa;

	return ((oslHostAddr)pAddr);
}

/*****************************************************************************/
/* osl_createHostAddrByName */
/*****************************************************************************/
oslHostAddr SAL_CALL osl_createHostAddrByName(rtl_uString *strHostname)
{
	if ((strHostname == NULL) || (strHostname->length == 0))
		return ((oslHostAddr)NULL);

	if (__osl_attemptSocketDialupImpl())
	{
		struct hostent *he;
		rtl_String     *Hostname= NULL;

		rtl_uString2String(
			&Hostname, strHostname->buffer, strHostname->length,
			RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS);

		he= gethostbyname (Hostname->buffer);

		rtl_string_release (Hostname);
		return __osl_hostentToHostAddr (he);
	}
	return ((oslHostAddr)NULL);
}

/*****************************************************************************/
/* osl_createHostAddrByAddr */
/*****************************************************************************/
oslHostAddr SAL_CALL osl_createHostAddrByAddr(const oslSocketAddr Addr)
{
	const struct sockaddr *pAddr= (const struct sockaddr *)Addr;
	OSL_ASSERT(pAddr);

	if (pAddr == NULL)
		return ((oslHostAddr)NULL);

	if (pAddr->sa_family == FAMILY_TO_NATIVE(osl_Socket_FamilyInet))
	{
		const struct sockaddr_in *sin= (const struct sockaddr_in *)pAddr;

		if (sin->sin_addr.s_addr == htonl(INADDR_ANY))
			return ((oslHostAddr)NULL);

		if (__osl_attemptSocketDialupImpl())
		{
			struct hostent *he;
			he= gethostbyaddr ((const sal_Char *)&(sin->sin_addr),
							   sizeof (sin->sin_addr),
							   sin->sin_family);
			return __osl_hostentToHostAddr (he);
		}
	}

	return ((oslHostAddr)NULL);
}

/*****************************************************************************/
/* osl_copyHostAddr */
/*****************************************************************************/
oslHostAddr SAL_CALL osl_copyHostAddr(const oslHostAddr Addr)
{
	oslHostAddrImpl *pAddr= (oslHostAddrImpl *)Addr;
	OSL_ASSERT(pAddr);

	if (pAddr)
		return osl_createHostAddr (pAddr->pHostName, pAddr->pSockAddr);
	else
		return ((oslHostAddr)NULL);
}

/*****************************************************************************/
/* osl_getHostnameOfHostAddr */
/*****************************************************************************/
void SAL_CALL osl_getHostnameOfHostAddr(
	const oslHostAddr Addr, rtl_uString **strHostname)
{
	const oslHostAddrImpl *pAddr= (const oslHostAddrImpl *)Addr;
	OSL_ASSERT(pAddr);

	if (pAddr)
		rtl_uString_assign (strHostname, pAddr->pHostName);
	else
		rtl_uString_new (strHostname);
}

/*****************************************************************************/
/* osl_getSocketAddrOfHostAddr */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_getSocketAddrOfHostAddr(const oslHostAddr Addr)
{
	const oslHostAddrImpl *pAddr= (const oslHostAddrImpl *)Addr;
	OSL_ASSERT(pAddr);

	if (pAddr)
		return (const oslSocketAddr)(pAddr->pSockAddr);
	else
		return NULL;
}

/*****************************************************************************/
/* osl_destroyHostAddr */
/*****************************************************************************/
void SAL_CALL osl_destroyHostAddr(oslHostAddr Addr)
{
	oslHostAddrImpl *pAddr = (oslHostAddrImpl *)Addr;

	if (pAddr)
	{
		if (pAddr->pHostName)
			rtl_uString_release (pAddr->pHostName);
		if (pAddr->pSockAddr)
			osl_destroySocketAddr (pAddr->pSockAddr);

		free (pAddr);
	}
}

/*****************************************************************************/
/* osl_getLocalHostname */
/*****************************************************************************/
oslSocketResult SAL_CALL osl_getLocalHostname (rtl_uString **strLocalHostname)
{
	static sal_Unicode LocalHostname[256] = L"";

	if (wcslen(LocalHostname) == 0)
	{
		sal_Char Host[256]= "";
		if (gethostname(Host, sizeof(Host)) == 0)
		{
			/* check if we have an FQDN */
	    	if (strchr(Host, '.') == NULL)
	        {
				oslHostAddrImpl *pAddr;
				rtl_uString     *hostName= NULL;
				
				rtl_string2UString(
					&hostName, Host, strlen(Host),
					RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS);

				/* no, determine it via dns */
				pAddr = (oslHostAddrImpl*)osl_createHostAddrByName(hostName);
				rtl_uString_release (hostName);

				if (pAddr && pAddr->pHostName)
					wcscpy(LocalHostname, pAddr->pHostName->buffer);
				else
					memset(LocalHostname, 0, sizeof(LocalHostname));

				osl_destroyHostAddr ((oslHostAddr)pAddr);
			}
		}
	}

	if (wcslen(LocalHostname) > 0)
	{
		rtl_uString_newFromStr (strLocalHostname, LocalHostname);
		return osl_Socket_Ok;
	}

	return osl_Socket_Error;
}

/*****************************************************************************/
/* osl_resolveHostname */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_resolveHostname(rtl_uString* strHostname)
{
	oslHostAddrImpl *pAddr=
		(oslHostAddrImpl *)osl_createHostAddrByName (strHostname);
	if (pAddr)
	{
		oslSocketAddr SockAddr = osl_copySocketAddr(pAddr->pSockAddr);

		osl_destroyHostAddr(pAddr); 

		return (SockAddr);
	}
	return ((oslSocketAddr)NULL);
}

/*****************************************************************************/
/* osl_getServicePort */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_getServicePort (
	rtl_uString* strServicename,
	rtl_uString* strProtocol)
{
	struct servent* ps;

	rtl_String *str_Servicename=NULL;
	rtl_String *str_Protocol=NULL;

	rtl_uString2String(
		&str_Servicename,
		rtl_uString_getStr(strServicename),
		rtl_uString_getLength(strServicename),
		RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS);
	rtl_uString2String(
		&str_Protocol,
		rtl_uString_getStr(strProtocol),
		rtl_uString_getLength(strProtocol),
		RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS);

	ps= getservbyname(
		rtl_string_getStr(str_Servicename),
		rtl_string_getStr(str_Protocol));

	rtl_string_release( str_Servicename );
	rtl_string_release( str_Protocol );

	if (ps != 0)
		return ntohs(ps->s_port);

	return OSL_INVALID_PORT;
}

/*****************************************************************************/
/* osl_destroySocketAddr */
/*****************************************************************************/
void SAL_CALL osl_destroySocketAddr(oslSocketAddr Addr)
{
	if(Addr) 
		free(Addr);
}

/*****************************************************************************/
/* osl_getFamilyOfSocketAddr */
/*****************************************************************************/
oslAddrFamily SAL_CALL osl_getFamilyOfSocketAddr(oslSocketAddr Addr)
{
	struct sockaddr* pAddr= (struct sockaddr*)Addr;
	OSL_ASSERT(pAddr);

	if (pAddr)
		return FAMILY_FROM_NATIVE(pAddr->sa_family);
	else
		return osl_Socket_FamilyInvalid;
}

/*****************************************************************************/
/* osl_getInetPortOfSocketAddr */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_getInetPortOfSocketAddr(oslSocketAddr Addr)
{
	struct sockaddr_in* pAddr= (struct sockaddr_in*)Addr;
	OSL_ASSERT(pAddr);
	
	if (pAddr && (pAddr->sin_family == FAMILY_TO_NATIVE(osl_Socket_FamilyInet)))
		return ntohs(pAddr->sin_port);
	else
		return OSL_INVALID_PORT;
}

/*****************************************************************************/
/* osl_setInetPortOfSocketAddr */
/*****************************************************************************/
sal_Bool SAL_CALL osl_setInetPortOfSocketAddr (
	oslSocketAddr Addr,
	sal_Int32     Port)
{
	struct sockaddr_in* pAddr= (struct sockaddr_in*)Addr;
	OSL_ASSERT(pAddr);

	if (pAddr == NULL)
		return sal_False;

	if (pAddr->sin_family != FAMILY_TO_NATIVE(osl_Socket_FamilyInet))
		return sal_False;

	pAddr->sin_port= htons((short)Port);
	return sal_True;
}

/*****************************************************************************/
/* osl_getHostnameOfSocketAddr */
/*****************************************************************************/
oslSocketResult SAL_CALL osl_getHostnameOfSocketAddr (
	oslSocketAddr   Addr,
	rtl_uString   **strHostName)
{
	oslHostAddrImpl *pAddr=
		(oslHostAddrImpl*)osl_createHostAddrByAddr (Addr);

	if (pAddr)
	{
		rtl_uString_newFromString(strHostName, pAddr->pHostName);

		osl_destroyHostAddr(pAddr);

		return osl_Socket_Ok;
	}

	return osl_Socket_Error;
}

/*****************************************************************************/
/* osl_getDottedInetAddrOfSocketAddr */
/*****************************************************************************/
oslSocketResult	SAL_CALL osl_getDottedInetAddrOfSocketAddr (
	oslSocketAddr   Addr,
	rtl_uString   **strDottedInetAddr)
{
	struct sockaddr_in *pAddr;
	sal_Char           *pDotted;

	pAddr= (struct sockaddr_in*)Addr;
	OSL_ASSERT(pAddr);

	if (pAddr == NULL)
		return osl_Socket_Error;

	if (pAddr->sin_family != FAMILY_TO_NATIVE(osl_Socket_FamilyInet))
		return osl_Socket_Error;

	pDotted = inet_ntoa (pAddr->sin_addr);
	rtl_string2UString(
		strDottedInetAddr, pDotted, strlen (pDotted),
		RTL_TEXTENCODING_UTF8, OUSTRING_TO_OSTRING_CVTFLAGS);

	return osl_Socket_Ok;
}

/*****************************************************************************/
/* osl_getIpxNetNumber  */
/*****************************************************************************/
oslSocketResult SAL_CALL osl_getIpxNetNumber (
	oslSocketAddr         Addr,
	oslSocketIpxNetNumber NetNumber)
{
	struct sockaddr_ipx* pAddr;

	pAddr= (struct sockaddr_ipx*)Addr;
	OSL_ASSERT(pAddr);
	
	if (pAddr && (pAddr->sa_family == FAMILY_TO_NATIVE(osl_Socket_FamilyIpx)))
	{
		memcpy(NetNumber, pAddr->sa_netnum, sizeof(NetNumber));

		return osl_Socket_Ok;
	}
	else
		return osl_Socket_Error;
}

/*****************************************************************************/
/* osl_getIpxNodeNumber  */
/*****************************************************************************/
oslSocketResult SAL_CALL osl_getIpxNodeNumber (
	oslSocketAddr          Addr,
	oslSocketIpxNodeNumber NodeNumber)
{
	struct sockaddr_ipx* pAddr;
	
	pAddr= (struct sockaddr_ipx*)Addr;
	OSL_ASSERT(pAddr);
	
	if (pAddr && (pAddr->sa_family == FAMILY_TO_NATIVE(osl_Socket_FamilyIpx)))
	{
		memcpy(NodeNumber, pAddr->sa_nodenum, sizeof(NodeNumber));

		return osl_Socket_Ok;
	}
	else
		return osl_Socket_Error;
}

/*****************************************************************************/
/* osl_getIpxSocketNumber  */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_getIpxSocketNumber(oslSocketAddr Addr)
{
	struct sockaddr_ipx* pAddr;
	
	pAddr= (struct sockaddr_ipx*)Addr;
	OSL_ASSERT(pAddr);
	
	if (pAddr && (pAddr->sa_family == FAMILY_TO_NATIVE(osl_Socket_FamilyIpx)))
		return pAddr->sa_socket;
	else
		return OSL_INVALID_IPX_SOCKET_NO;
}

/*****************************************************************************/
/* osl_createSocket  */
/*****************************************************************************/
oslSocket SAL_CALL osl_createSocket (
	oslAddrFamily Family, 
	oslSocketType Type, 
	oslProtocol   Protocol)
{
	oslSocketImpl* pSockImpl;

	OSL_TRACE("Creating Socket. Family: %d Type: %d Protocol: %d\n",
			  FAMILY_TO_NATIVE(Family),
			  TYPE_TO_NATIVE(Type),
			  PROTOCOL_TO_NATIVE(Protocol));

	/* alloc memory */
	pSockImpl= __osl_createSocketImpl(0);
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL)
		return ((oslSocket)NULL);
	
	/* create socket */
	pSockImpl->m_Socket= socket(FAMILY_TO_NATIVE(Family), 
								TYPE_TO_NATIVE(Type), 
								PROTOCOL_TO_NATIVE(Protocol));

	/* creation failed => free memory */
	if(pSockImpl->m_Socket == OSL_INVALID_SOCKET)
	{
		__osl_destroySocketImpl(pSockImpl);
		pSockImpl= 0;
	}
	else
	{
		pSockImpl->m_Flags			= 0;
		pSockImpl->m_CloseCallback	= NULL;
		pSockImpl->m_CallbackArg	= NULL;
	}

	return ((oslSocket)pSockImpl);
}

/*****************************************************************************/
/* osl_copySocket  */
/*****************************************************************************/
oslSocket SAL_CALL osl_copySocket(oslSocket Socket)
{
	oslSocketImpl* pSockImpl;
	oslSocketImpl* pParamSockImpl;

	pParamSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pParamSockImpl);

	if (pParamSockImpl == NULL) /* ENOTSOCK */
		return ((oslSocket)NULL);

	/* alloc memory */
	pSockImpl= __osl_createSocketImpl(0);
	OSL_ASSERT(pSockImpl);
	
	if (pSockImpl)
	{
		/* copy socket */
		pSockImpl->m_Socket			= pParamSockImpl->m_Socket;
		pSockImpl->m_Flags			= 0;
		pSockImpl->m_CloseCallback	= NULL;
		pSockImpl->m_CallbackArg	= NULL;
	}

	return ((oslSocket)pSockImpl);
}

/*****************************************************************************/
/* osl_destroySocket  */
/*****************************************************************************/
void SAL_CALL osl_destroySocket(oslSocket Socket)
{
	oslSocketImpl* pSockImpl;

	/* socket already invalid */
	if(Socket==0) 
		return;

	pSockImpl= (oslSocketImpl*)Socket;

	/* close */
	closesocket(pSockImpl->m_Socket);

	/* registrierten Callback ausfuehren */
	if (pSockImpl->m_CloseCallback != NULL)
	{
		pSockImpl->m_CloseCallback(pSockImpl->m_CallbackArg);
	}

	/* free memory */
	__osl_destroySocketImpl(pSockImpl);
}

/*****************************************************************************/
/* osl_closeSocket  */
/*****************************************************************************/
void SAL_CALL osl_closeSocket(oslSocket Socket)
{

	oslSocketImpl* pSockImpl;

	/* socket already invalid */
	if(Socket==0) 
		return;

	pSockImpl= (oslSocketImpl*)Socket;

	/* close */
	closesocket(pSockImpl->m_Socket);

	pSockImpl->m_Socket = OSL_INVALID_SOCKET;

	/* registrierten Callback ausfuehren */
	if (pSockImpl->m_CloseCallback != NULL)
	{
		pSockImpl->m_CloseCallback(pSockImpl->m_CallbackArg);
	}
}

/*****************************************************************************/
/* osl_getLocalAddrOfSocket  */
/* Note that I rely on the fact that oslSocketAddr and struct sockaddr */
/* are the same! I don't like it very much but see no other easy way */
/* to conceal the struct sockaddr from the eyes of the user. */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_getLocalAddrOfSocket(oslSocket Socket)
{
	oslSocketImpl*  pSockImpl;
	struct sockaddr Addr;
	int             AddrLen;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return ((oslSocketAddr)NULL);

	AddrLen= sizeof(struct sockaddr);
  
	if (getsockname(pSockImpl->m_Socket, &Addr, &AddrLen) == OSL_SOCKET_ERROR)
		return ((oslSocketAddr)NULL);

	return osl_copySocketAddr((oslSocketAddr)&Addr);
}

/*****************************************************************************/
/* osl_getPeerAddrOfSocket  */
/*****************************************************************************/
oslSocketAddr SAL_CALL osl_getPeerAddrOfSocket(oslSocket Socket)
{
	oslSocketImpl*  pSockImpl;
	struct sockaddr Addr;
	int             AddrLen;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return ((oslSocketAddr)NULL);

	AddrLen= sizeof(struct sockaddr);
  
	if (getpeername(pSockImpl->m_Socket, &Addr, &AddrLen) == OSL_SOCKET_ERROR)
		return ((oslSocketAddr)NULL);

	return osl_copySocketAddr((oslSocketAddr)&Addr);
}

/*****************************************************************************/
/* osl_bindAddrToSocket  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_bindAddrToSocket (
	oslSocket     Socket,
	oslSocketAddr Addr)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	return (bind(pSockImpl->m_Socket,
				 (struct sockaddr*)Addr,
				 sizeof(struct sockaddr)) != OSL_SOCKET_ERROR);
}

/*****************************************************************************/
/* osl_connectSocketTo  */
/*****************************************************************************/
oslSocketResult SAL_CALL osl_connectSocketTo (
	oslSocket        Socket,
	oslSocketAddr    Addr,
	const TimeValue* pTimeout)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return osl_Socket_Error;

	OSL_ASSERT(Addr);

	if (Addr == NULL) /* EDESTADDRREQ */
		return osl_Socket_Error;

	if (!__osl_attemptSocketDialupImpl()) /* ENETDOWN */
		return osl_Socket_Error;

	if (pTimeout == NULL)
	{
		if(connect(pSockImpl->m_Socket, 
				   (struct sockaddr*)Addr, 
					sizeof(struct sockaddr)) == OSL_SOCKET_ERROR)
			return osl_Socket_Error;
		else
			return osl_Socket_Ok;
	}
	else
	{
		fd_set          fds;
		int	            error;
	    struct timeval  tv;
		unsigned long   Param;
		oslSocketResult Result= osl_Socket_Ok;

		if (pSockImpl->m_Flags & OSL_SOCKET_FLAGS_NONBLOCKING) 
	    {
			if (connect(pSockImpl->m_Socket, 
			    	    (struct sockaddr*)Addr, 
						sizeof(struct sockaddr)) == OSL_SOCKET_ERROR)
			{
			    switch (WSAGetLastError())
			    {
					case WSAEWOULDBLOCK:
					case WSAEINPROGRESS:
						return osl_Socket_InProgress;
										
					default:
						return osl_Socket_Error;
				}
			}
			else
				return osl_Socket_Ok;
		}

		/* set socket temporarily to non-blocking */
		Param= 1;
		OSL_VERIFY(ioctlsocket(
			pSockImpl->m_Socket, FIONBIO, &Param) != OSL_SOCKET_ERROR);
		    
		/* initiate connect */
		if (connect(pSockImpl->m_Socket, 
		     	    (struct sockaddr*)Addr, 
				    sizeof(struct sockaddr)) != OSL_SOCKET_ERROR)
		{
		   /* immediate connection */

			Param= 0;
			ioctlsocket(pSockImpl->m_Socket, FIONBIO, &Param);

			return osl_Socket_Ok;
	    } 
		else 
		{
			error = WSAGetLastError();

		    /* really an error or just delayed? */
		    if (error != WSAEWOULDBLOCK && error != WSAEINPROGRESS)
		    {
			     Param= 0;
			     ioctlsocket(pSockImpl->m_Socket, FIONBIO, &Param);

			     return osl_Socket_Error;
		    }
		}
		
		/* prepare select set for socket  */
		FD_ZERO(&fds);
		FD_SET(pSockImpl->m_Socket, &fds); 

		/* divide milliseconds into seconds and microseconds */
		tv.tv_sec=	pTimeout->Seconds;
		tv.tv_usec=	pTimeout->Nanosec / 1000L;	

		/* select */
	    error= select(pSockImpl->m_Socket+1,					
		 		      0,
					  &fds,
					  0,
					  &tv);	

		if (error > 0)  /* connected */
		{
			OSL_POSTCOND(
				FD_ISSET(pSockImpl->m_Socket, &fds),
				"osl_connectSocketTo(): select returned but socket not set\n");

			Result= osl_Socket_Ok;

		} 
		else if(error < 0)  /* error */
		{
		    /* errno == EBADF: most probably interrupted by close() */
		    if(WSAGetLastError() == WSAEBADF)
			{
			    /* do not access pSockImpl because it is about to be or */
			    /* already destroyed */
			    return osl_Socket_Interrupted;
			} 
			else
			    Result= osl_Socket_Error;

		} 
		else    /* timeout */
			Result= osl_Socket_TimedOut;


		/* clean up */
		Param= 0;
		ioctlsocket(pSockImpl->m_Socket, FIONBIO, &Param);

		return Result;
	}
}

/*****************************************************************************/
/* osl_listenOnSocket  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_listenOnSocket (
	oslSocket  Socket,
	sal_Int32  MaxPendingConnections)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	return (listen(pSockImpl->m_Socket, 
				   MaxPendingConnections == -1 ? 
				   SOMAXCONN :
				   MaxPendingConnections) != OSL_SOCKET_ERROR);
}

/*****************************************************************************/
/* osl_acceptConnectionOnSocket  */
/*****************************************************************************/
oslSocket SAL_CALL osl_acceptConnectionOnSocket (
	oslSocket      Socket,
	oslSocketAddr* pAddr)
{
	oslSocketImpl*  pSockImpl;
	struct sockaddr Addr;
	SOCKET          Connection;
	oslSocketImpl*  pConnectionSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return ((oslSocket)NULL);

	if(pAddr) 
	{
		int AddrLen= sizeof(struct sockaddr);

		/* user wants to know peer Addr */
		Connection=	accept(pSockImpl->m_Socket, 
						   (struct sockaddr*)&Addr, 
							&AddrLen);
		OSL_ASSERT(AddrLen == sizeof(struct sockaddr));

		if(Connection != OSL_SOCKET_ERROR)
			*pAddr= osl_copySocketAddr((oslSocketAddr)&Addr);
		else
			*pAddr = NULL;
	}
	else 
	{
		/* user is not interested in peer-addr */
		Connection=	accept(pSockImpl->m_Socket, 0, 0);
	}

	/* accept failed? */
	if(Connection == OSL_SOCKET_ERROR)
		return ((oslSocket)NULL);

	/* alloc memory */
	pConnectionSockImpl= __osl_createSocketImpl(0);

	pConnectionSockImpl->m_Socket			= Connection;
	pConnectionSockImpl->m_Flags			= 0;
	pConnectionSockImpl->m_CloseCallback	= NULL;
	pConnectionSockImpl->m_CallbackArg		= NULL;

	return ((oslSocket)pConnectionSockImpl);
}

/*****************************************************************************/
/* osl_receiveSocket  */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_receiveSocket (
	oslSocket        Socket,
	void*            pBuffer,
	sal_uInt32       BytesToRead,
	oslSocketMsgFlag Flag)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return osl_Socket_Error;

	return recv(pSockImpl->m_Socket, 
				(sal_Char*)pBuffer, 
				BytesToRead, 
				MSG_FLAG_TO_NATIVE(Flag));
}

/*****************************************************************************/
/* osl_receiveFromSocket  */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_receiveFromSocket (
	oslSocket        Socket, 
	oslSocketAddr    SenderAddr,
	void*            pBuffer,
	sal_uInt32       BufferSize,
	oslSocketMsgFlag Flag)
{
	oslSocketImpl* pSockImpl;
	int AddrLen= SenderAddr == 0 ? 0 : sizeof(struct sockaddr);

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return osl_Socket_Error;

	return recvfrom(pSockImpl->m_Socket, 
					 (sal_Char*)pBuffer, 
					 BufferSize, 
					 MSG_FLAG_TO_NATIVE(Flag),
					 (struct sockaddr*)SenderAddr,
					 &AddrLen);
}

/*****************************************************************************/
/* osl_sendSocket  */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_sendSocket (
	oslSocket        Socket, 
	const void*      pBuffer,
	sal_uInt32       BytesToSend,
	oslSocketMsgFlag Flag)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return osl_Socket_Error;

	return send(pSockImpl->m_Socket, 
				(sal_Char*)pBuffer, 
				BytesToSend, 
				MSG_FLAG_TO_NATIVE(Flag));
}

/*****************************************************************************/
/* osl_sendToSocket  */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_sendToSocket (
	oslSocket        Socket, 
	oslSocketAddr    ReceiverAddr,
	const void*      pBuffer,
	sal_uInt32       BytesToSend,
	oslSocketMsgFlag Flag)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return osl_Socket_Error;

	/* ReceiverAddr might be 0 when used on a connected socket. */
	/* Then sendto should behave like send. */

	return sendto(pSockImpl->m_Socket, 
				  (sal_Char*)pBuffer, 
				  BytesToSend, 
				  MSG_FLAG_TO_NATIVE(Flag),
				  (struct sockaddr*)ReceiverAddr,
				  ReceiverAddr == 0 ? 0 : sizeof(struct sockaddr));
}

/*****************************************************************************/
/* osl_isReceiveReady  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_isReceiveReady (
	oslSocket        Socket,
	const TimeValue* pTimeout)
{
	fd_set         fds;
	struct timeval tv;
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	FD_ZERO(&fds);
	FD_SET(pSockImpl->m_Socket, &fds);

	if (pTimeout)
	{
		tv.tv_sec  = pTimeout->Seconds;
		tv.tv_usec = pTimeout->Nanosec / 1000L;
	}

	return (select(pSockImpl->m_Socket + 1,		/* no of sockets to monitor */
				   &fds,						/* check read operations */
				   0,							/* check write ops */
				   0,							/* ckeck for OOB */
				   (pTimeout) ? &tv : 0)==1);	/* use timeout? */
}

/*****************************************************************************/
/* osl_isSendReady  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_isSendReady (
	oslSocket        Socket,
	const TimeValue* pTimeout)
{
	fd_set         fds;
	struct timeval tv;
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	FD_ZERO(&fds);
	FD_SET(pSockImpl->m_Socket, &fds);
  
	if (pTimeout)
	{
		tv.tv_sec  = pTimeout->Seconds;
		tv.tv_usec = pTimeout->Nanosec / 1000L;
	}
  
	return (select(pSockImpl->m_Socket + 1,		/* no of sockets to monitor */
				   0,							/* check read operations */
				   &fds,						/* check write ops */
				   0,							/* ckeck for OOB */
				   (pTimeout) ? &tv : 0)==1);	/* use timeout? */
}

/*****************************************************************************/
/* osl_isExceptionPending  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_isExceptionPending (
	oslSocket        Socket,
	const TimeValue* pTimeout)
{
	fd_set         fds;
	struct timeval tv;
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	FD_ZERO(&fds);
	FD_SET(pSockImpl->m_Socket, &fds);
  
	if (pTimeout)
	{
		tv.tv_sec  = pTimeout->Seconds;
		tv.tv_usec = pTimeout->Nanosec / 1000L;
	}
  
	return (select(pSockImpl->m_Socket + 1,	    /* no of sockets to monitor */
				   0,							/* check read operations */
				   0,							/* check write ops */
				   &fds,						/* ckeck for OOB */
				   (pTimeout) ? &tv : 0)==1);	/* use timeout? */
}

/*****************************************************************************/
/* osl_shutdownSocket  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_shutdownSocket (
	oslSocket          Socket, 
	oslSocketDirection Direction)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	return (shutdown(pSockImpl->m_Socket, DIRECTION_TO_NATIVE(Direction))==0);
}

/*****************************************************************************/
/* osl_getSocketOption  */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_getSocketOption (
	oslSocket            Socket,
	oslSocketOptionLevel Level,
	oslSocketOption      Option,
	void*                pBuffer, 
	sal_uInt32           BufferLen)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return osl_Socket_Error;

	if (getsockopt(pSockImpl->m_Socket, 
	     		   OPTION_LEVEL_TO_NATIVE(Level), 
				   OPTION_TO_NATIVE(Option), 
				   (sal_Char*)pBuffer, 
				   &BufferLen) == -1)
	{
		(sal_Int32)BufferLen = -1;
	}

	return (sal_Int32)BufferLen;
}

/*****************************************************************************/
/* osl_setSocketOption  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_setSocketOption (
	oslSocket            Socket,
	oslSocketOptionLevel Level,
	oslSocketOption      Option,
	void*                pBuffer, 
	sal_uInt32           BufferLen)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	return(setsockopt(pSockImpl->m_Socket, 
					  OPTION_LEVEL_TO_NATIVE(Level), 
					  OPTION_TO_NATIVE(Option), 
					  (sal_Char*)pBuffer, 
					  BufferLen) == 0);
}

/*****************************************************************************/
/* osl_enableNonBlockingMode  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_enableNonBlockingMode (
	oslSocket Socket,
	sal_Bool  On)
{
	oslSocketImpl* pSockImpl;
	unsigned long  Param= On ? 1 : 0;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	pSockImpl->m_Flags = Param ? 
		(pSockImpl->m_Flags |  OSL_SOCKET_FLAGS_NONBLOCKING) :
		(pSockImpl->m_Flags & ~OSL_SOCKET_FLAGS_NONBLOCKING) ;

	return (
		ioctlsocket(pSockImpl->m_Socket, FIONBIO, &Param) != OSL_SOCKET_ERROR);
}

/*****************************************************************************/
/* osl_isNonBlockingMode  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_isNonBlockingMode(oslSocket Socket)
{
	oslSocketImpl* pSockImpl;

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return sal_False;

	return (pSockImpl->m_Flags & OSL_SOCKET_FLAGS_NONBLOCKING);
}

/*****************************************************************************/
/* osl_getSocketType  */
/*****************************************************************************/
oslSocketType SAL_CALL osl_getSocketType(oslSocket Socket)
{
	oslSocketImpl* pSockImpl;
	int            Type=0;
	int            TypeSize= sizeof(Type);

	pSockImpl= (oslSocketImpl*)Socket;
	OSL_ASSERT(pSockImpl);

	if (pSockImpl == NULL) /* ENOTSOCK */
		return osl_Socket_TypeInvalid;

	if(getsockopt(pSockImpl->m_Socket, 
				  OPTION_LEVEL_TO_NATIVE(osl_Socket_LevelSocket), 
				  OPTION_TO_NATIVE(osl_Socket_OptionType), 
				  (sal_Char *)&Type, 
				  &TypeSize) == -1)
	{
		/* error */
		return osl_Socket_TypeInvalid;
	}
  
	return TYPE_FROM_NATIVE(Type);
}

/*****************************************************************************/
/* osl_getLastSocketErrorDescription  */
/*****************************************************************************/
void SAL_CALL osl_getLastSocketErrorDescription (
	oslSocket  Socket,
	rtl_uString	**strError)
{
	int error;

	switch(error = WSAGetLastError())
	{
		case WSAENOTSOCK:
			rtl_uString_newFromStr (strError, L"WSAENOTSOCK, Socket operation on non-socket. A socket created in one process is used by another process.");
            break;

        case WSAEDESTADDRREQ:
			rtl_uString_newFromStr (strError, L"WSAEDESTADDRREQ, Destination Addr required");
            break;

        case WSAEMSGSIZE:
            rtl_uString_newFromStr (strError, L"WSAEMSGSIZE, Message too long");
            break;

        case WSAEPROTOTYPE:
			rtl_uString_newFromStr (strError, L"WSAEPROTOTYPE, Protocol wrong type for socket");
            break;

        case WSAENOPROTOOPT:
			rtl_uString_newFromStr (strError, L"WSAENOPROTOOPT, Protocol not available");
            break;

        case WSAEPROTONOSUPPORT:
			rtl_uString_newFromStr (strError, L"WSAEPROTONOSUPPORT, Protocol not supported");
            break;

        case WSAESOCKTNOSUPPORT:
			rtl_uString_newFromStr (strError, L"WSAESOCKTNOSUPPORT, Socket type not supported");
            break;

        case WSAEOPNOTSUPP:
			rtl_uString_newFromStr (strError, L"WSAEOPNOTSUPP, Operation not supported on socket");
            break;

        case WSAEPFNOSUPPORT:
			rtl_uString_newFromStr (strError, L"WSAEPFNOSUPPORT, Protocol family not supported");
            break;

        case WSAEAFNOSUPPORT:
			rtl_uString_newFromStr (strError, L"WSEAFNOSUPPORT, Addr family not supported by protocol family");
            break;

        case WSAEADDRINUSE:
			rtl_uString_newFromStr (strError, L"WSAEADDRINUSE, Triggered by bind() because a process went down without closing a socket.");
            break;

        case WSAEADDRNOTAVAIL:
			rtl_uString_newFromStr (strError, L"WSAEADDRNOTAVAIL, Can't assign requested Addr");
            break;

        case WSAENETDOWN:
			rtl_uString_newFromStr (strError, L"WSAENETDOWN, Network is down");
            break;

        case WSAENETUNREACH:
			rtl_uString_newFromStr (strError, L"WSAENETUNREACH, Network is unreachable");
            break;

        case WSAENETRESET:
			rtl_uString_newFromStr (strError, L"WSAENETRESET, Network dropped connection or reset");
            break;

        case WSAECONNABORTED:
			rtl_uString_newFromStr (strError, L"WSAECONNABORTED, Software caused connection abort");
            break;

        case WSAECONNRESET:
			rtl_uString_newFromStr (strError, L"WSAECONNRESET, Connection reset by peer");
            break;

        case WSAENOBUFS:
			rtl_uString_newFromStr (strError, L"WSAENOBUFS, No buffer space available.");
            break;

        case WSAEISCONN:
			rtl_uString_newFromStr (strError, L"WSAEISCONN, Socket is already connected");
            break;

        case WSAENOTCONN:
			rtl_uString_newFromStr (strError, L"WSAENOTCONN, Socket is not connected");
            break;

        case WSAESHUTDOWN:
			rtl_uString_newFromStr (strError, L"WSAESHUTDOWN, Can't send after socket shutdown");
            break;

        case WSAETIMEDOUT:
			rtl_uString_newFromStr (strError, L"WSAETIMEDOUT, Connection timed out");
            break;

        case WSAECONNREFUSED:
			rtl_uString_newFromStr (strError, L"WSAECONNREFUSED, Connection refused");
            break;

        case WSAEHOSTDOWN:
			rtl_uString_newFromStr (strError, L"WSAEHOSTDOWN, Networking subsystem not started");
            break;

        case WSAEHOSTUNREACH:
			rtl_uString_newFromStr (strError, L"WSAEHOSTUNREACH, No route to host");
            break;

        case WSAEWOULDBLOCK:
			rtl_uString_newFromStr (strError, L"WSAEWOULDBLOCK, Operation would block");
            break;

        case WSAEINPROGRESS:
			rtl_uString_newFromStr (strError, L"WSAEINPROGRESS, Operation now in progress");
            break;

        case WSAEALREADY:
			rtl_uString_newFromStr (strError, L"WSAEALREADY, Operation already in progress");
            break;

        case WSAEINTR:
			rtl_uString_newFromStr (strError, L"WSAEALREADY, Operation was interrupted");
            break;

        case WSAEBADF:
			rtl_uString_newFromStr (strError, L"WSAEBADF, Bad file number");
            break;

        case WSAEACCES:
			rtl_uString_newFromStr (strError, L"WSAEACCES, Access is denied");
            break;

        case WSAEFAULT:
			rtl_uString_newFromStr (strError, L"WSAEFAULT, Bad memory Addr");
            break;

        case WSAEINVAL:
			rtl_uString_newFromStr (strError, L"WSAEINVAL, The socket has not been bound with bind() or is already connected");
            break;

        case WSAEMFILE:
			rtl_uString_newFromStr (strError, L"WSAEMFILE, No more file descriptors are available");
            break;

        case WSAETOOMANYREFS:
			rtl_uString_newFromStr (strError, L"WSAETOOMANYREFS, Undocumented WinSock error");
            break;

        case WSAENAMETOOLONG:
			rtl_uString_newFromStr (strError, L"WSAENAMETOOLONG, Undocumented WinSock error");
            break;

        case WSAENOTEMPTY:
			rtl_uString_newFromStr (strError, L"WSAENOTEMPTY, Undocumented WinSock error");
            break;

        case WSAEPROCLIM:
			rtl_uString_newFromStr (strError, L"WSAEPROCLIM, Undocumented WinSock error");
            break;

        case WSAEUSERS:
			rtl_uString_newFromStr (strError, L"WSAEUSERS, Undocumented WinSock error");
            break;

        case WSAEDQUOT:
			rtl_uString_newFromStr (strError, L"WSAEDQUOT, Undocumented WinSock error");
            break;

        case WSAESTALE:
			rtl_uString_newFromStr (strError, L"WSAESTALE, Undocumented WinSock error");
            break;

        case WSAEREMOTE:
			rtl_uString_newFromStr (strError, L"WSAEREMOTE, Undocumented WinSock error");
            break;

        case WSAEDISCON:
			rtl_uString_newFromStr (strError, L"WSAEDISCON, Circuit was gracefully terminated");
            break;

        case WSASYSNOTREADY:
			rtl_uString_newFromStr (strError, L"WSASYSNOTREADY, The underlying network subsystem is not ready for network communication");
            break;

        case WSAVERNOTSUPPORTED:
			rtl_uString_newFromStr (strError, L"WSAVERNOTSUPPORTED, The version of Windows Sockets API support requested is not provided by this particular Windows Sockets implementation");
            break;

        case WSANOTINITIALISED:
			rtl_uString_newFromStr (strError, L"WSANOTINITIALISED, WSAStartup() has not been called");
            break;

        case WSAHOST_NOT_FOUND:
			rtl_uString_newFromStr (strError, L"WSAHOST_NOT_FOUND, Authoritative answer host not found");
            break;

        case WSATRY_AGAIN:
			rtl_uString_newFromStr (strError, L"WSATRY_AGAIN, Non-authoritative answer host not found or SERVERFAIL");
            break;

        case WSANO_RECOVERY:
			rtl_uString_newFromStr (strError, L"WSANO_RECOVERY, Non recoverable errors, FORMERR, REFUSED, NOTIMP");
            break;

        case WSANO_DATA:
			rtl_uString_newFromStr (strError, L"WSANO_DATA or WSANO_ADDRESS, Valid name, no data record of requested type");
            break;

		default:
		{
			sal_Unicode message[128];

            wsprintfW((sal_Unicode *)message, (sal_Unicode *)L"Unknown WinSock Error Number %d", error);
			rtl_uString_newFromStr (strError, message);
        }

        return;

	}
}

/*****************************************************************************/
/* osl_getLastSocketError  */
/*****************************************************************************/
oslSocketError SAL_CALL osl_getLastSocketError(oslSocket Socket)
{
	return ERROR_FROM_NATIVE(WSAGetLastError());
}

/*****************************************************************************/
/* SocketSet                                                                 */
/*****************************************************************************/
typedef struct _TSocketSetImpl
{
	fd_set	m_Set;			/* the set of descriptors */

} TSocketSetImpl;

/*****************************************************************************/
/* osl_createSocketSet  */
/*****************************************************************************/
oslSocketSet SAL_CALL osl_createSocketSet()
{
	TSocketSetImpl* pSet;

	pSet= malloc(sizeof(TSocketSetImpl));

	OSL_ASSERT(pSet);

	if(pSet)
	{
		FD_ZERO(&pSet->m_Set);
	}

	return (oslSocketSet)pSet;
}

/*****************************************************************************/
/* osl_destroySocketSet  */
/*****************************************************************************/
void SAL_CALL osl_destroySocketSet (oslSocketSet Set)
{
	if(Set)
		free(Set);
}

/*****************************************************************************/
/* osl_clearSocketSet  */
/*****************************************************************************/
void SAL_CALL osl_clearSocketSet (oslSocketSet Set)
{
	TSocketSetImpl* pSet;

	pSet= (TSocketSetImpl*)Set;
	OSL_ASSERT(pSet);

	if (pSet)
		FD_ZERO(&pSet->m_Set);
}

/*****************************************************************************/
/* osl_addToSocketSet  */
/*****************************************************************************/
void SAL_CALL osl_addToSocketSet (
	oslSocketSet Set,
	oslSocket    Socket)
{
	TSocketSetImpl* pSet;
	oslSocketImpl*  pSockImpl;

	pSet= (TSocketSetImpl*)Set;
	pSockImpl= (oslSocketImpl*)Socket;
	
	OSL_ASSERT(pSet);
	OSL_ASSERT(pSockImpl);

	if (pSet && pSockImpl)
		FD_SET(pSockImpl->m_Socket, &pSet->m_Set);
}

/*****************************************************************************/
/* osl_removeFromSocketSet  */
/*****************************************************************************/
void SAL_CALL osl_removeFromSocketSet (
	oslSocketSet Set,
	oslSocket    Socket)
{
	TSocketSetImpl* pSet;
	oslSocketImpl*  pSockImpl;

	pSet= (TSocketSetImpl*)Set;
	pSockImpl= (oslSocketImpl*)Socket;

	OSL_ASSERT(pSet);
	OSL_ASSERT(pSockImpl);

	if (pSet && pSockImpl)
		FD_CLR(pSockImpl->m_Socket, &pSet->m_Set);
}

/*****************************************************************************/
/* osl_isInSocketSet  */
/*****************************************************************************/
sal_Bool SAL_CALL osl_isInSocketSet (
	oslSocketSet Set,
	oslSocket    Socket)
{
	TSocketSetImpl* pSet;
	oslSocketImpl*  pSockImpl;

	pSet= (TSocketSetImpl*)Set;
	pSockImpl= (oslSocketImpl*)Socket;

	OSL_ASSERT(pSet);
	OSL_ASSERT(pSockImpl);

	if (pSet && pSockImpl)
		return (FD_ISSET(pSockImpl->m_Socket, &pSet->m_Set) != 0);
	else
		return sal_False;
}

/*****************************************************************************/
/* osl_demultiplexSocketEvents  */
/*****************************************************************************/
sal_Int32 SAL_CALL osl_demultiplexSocketEvents (
	oslSocketSet IncomingSet,
	oslSocketSet OutgoingSet,
	oslSocketSet OutOfBandSet,
	const TimeValue* pTimeout)
{
	int             MaxHandle= 0;
	struct timeval  tv;
	TSocketSetImpl* pInSet;
	TSocketSetImpl* pOutSet;
	TSocketSetImpl* pOOBSet;

	if(pTimeout)
	{
		/* divide milliseconds into seconds and microseconds */
		tv.tv_sec  = pTimeout->Seconds;
		tv.tv_usec = pTimeout->Nanosec / 1000L;	
	}

	/* map opaque data to impl-types */
	pInSet= (TSocketSetImpl*)IncomingSet;
	pOutSet= (TSocketSetImpl*)OutgoingSet;
	pOOBSet= (TSocketSetImpl*)OutOfBandSet;
		
	return select(MaxHandle,				/* redundant in WIN32 */
				  pInSet ? &pInSet->m_Set : 0,
				  pOutSet ? &pOutSet->m_Set : 0,
				  pOOBSet ? &pOOBSet->m_Set : 0,
				  pTimeout ? &tv : 0);
}

/*************************************************************************
*
*    $Log: socket.c,v $
*    Revision 1.1.1.1  2000/09/18 15:17:23  hr
*    initial import
*
*    Revision 1.41  2000/09/18 14:29:02  willem.vandorp
*    OpenOffice header added.
*
*    Revision 1.40  2000/06/27 14:54:18  mhu
*    Added function osl_createInetBroadcastAddr().
*
*    Revision 1.39  2000/05/08 10:32:24  mhu
*    Fixed conversion to Unicode: internal rtl_String* initialization.
*
*    Revision 1.38  2000/03/31 16:24:43  rs
*    UNICODE-Changes
*
*    Revision 1.37  2000/03/06 21:07:30  mhu
*    Changed __osl_attemptSocketDialupImpl() to report ENETDOWN upon failure.
*    Adapted calling functions to propagate this failure (#67521#).
*
*    Revision 1.36  1999/10/28 20:04:03  mhu
*    Fixed osl_isInSocketSet() to convert FD_ISSET return into sal_Bool.
*
*    Revision 1.35  1999/10/27 15:13:37  mfe
*    Change of Copyright, removed compiler warnings, code clean up, ...
*
*    Revision 1.34  1999/10/11 12:50:27  rt
*    gcc for windows: added defines for shutdown
*
*    Revision 1.33  1999/07/21 16:22:28  ts
*    #67730# changed channel maneger api
*
*    Revision 1.32  1999/03/04 09:58:38  rh
*    #62583 compare socketaddr
*
*    Revision 1.31  1999/02/13 18:18:17  mhu
*    Added support for modem dialup handling (oslSocketDialupImpl) (#49995#).
*
*    Revision 1.30  1999/01/20 19:09:50  jsc
*    #61011# Typumstellung
*
*    Revision 1.29  1998/12/17 19:56:45  rh
*    #60480# osl_getLocalHostname fails if machine name ist already FQDN
*
*    Revision 1.28  1998/11/27 08:27:37  rh
*    #59877# osl_getHostname returns FDQN
*
*    Revision 1.27  1998/11/05 11:13:18  rh
*    #59037# Thread local storage implementation
*
*    Revision 1.26  1998/03/13 15:07:45  rh
*    Cleanup of enum chaos and implemntation of pipes
*
*    Revision 1.25  1998/03/06 15:42:12  rh
*    Added signal handling
*
*    Revision 1.24  1998/02/16 19:35:00  rh
*    Cleanup of ports, integration of sal_uInt32, features for process
*
*    Revision 1.23  1997/12/16 17:53:51  mhu
*    Fixed osl_getDottedInetAddrOfSocketAddr()
*    not to return osl_Socket_Error upon INADDR_ANY.
*
*    Revision 1.22  1997/12/15 09:56:40  ts
*    bugfix(d-46087): osl_acceptConnection() fehlende Memberinitalisierung
*    (m_closeCallback)
*
*    Revision 1.21  1997/12/04 18:40:31  ts
*    osl_createSocket(): keine Initaliserung der member, wenn der socket
*    nicht valid ist
*
*    Revision 1.20  1997/11/28 15:34:42  ts
*    bugfix: bei weitergereichten Sockets wird der kindprozess solange
*    blockiert, bis der vaterprozess diese Sockets closed
*
*    Revision 1.19  1997/10/22 10:15:48  jsc
*    osl_getHostnameOfSocketAddr bug entfernt
*		 
*    Revision 1.18  1997/10/16 16:33:25  rh
*    *** empty log message ***
*
*    Revision 1.17  1997/10/16 16:12:18  rh
*    *** empty log message ***
*
*    Revision 1.16  1997/10/15 10:29:25  mhu
*    Added function: sal_Bool osl_getLocalHostname (
*    rtl_uString *pBuffer, int nBufLen)
*
*    Revision 1.15  1997/09/23 16:49:31  mh
*    DEBUG
*
*    Revision 1.14  1997/09/15 10:55:43  mhu
*    Added function osl_copyHostAddr (const oslHostAddr Addr).
*    Fixed osl_getDottedInetAddrOfSocketAddr() to return NULL for INADDR_ANY
*    (as it did before).
*
*    Revision 1.13  1997/09/08 17:19:37  mhu
*    Added support for domain name resolution (oslHostAddr).
*    Improved assertion behaviour: assert but don't crash.
*
*    Revision 1.12  1997/09/04 09:51:46  rh
*    Blockmode
*
*    Revision 1.11  1997/08/26 09:30:53  ts
*    bind() bugfix
*
*    Revision 1.10  1997/08/23 20:36:48  rh
*    return value for name resolving
*
*    Revision 1.9  1997/08/22 14:47:45  rh
*    TimeValue inserted
*
*    Revision 1.8  1997/07/31 18:28:56  rh
*    Standard types adapted
*
*    Revision 1.7  1997/07/15 19:02:05  rh
*    system.h inserted
*
*    Revision 1.6  1997/07/02 10:54:02  rh
*    Anpassung fuer Diagnose Macros
*
*    Revision 1.5  1997/06/27 13:03:02  bho
*    *** empty log message ***
*
*    Revision 1.4  1997/06/27 12:57:19  bho
*    *** empty log message ***
*
*    Revision 1.3  1997/06/25 10:32:09  bho
*    changes in w32 sockets
*
*    Revision 1.2  1997/06/25 08:44:53  bho
*    !changed implementation of sems to SYS V, because POSIX Sems are not
*    supported yet by solaris.
*
*    Revision 1.1  1997/06/19 13:10:43  bho
*    first version of OSL.
*
*************************************************************************/

